#ifndef __UTIL_H__
#define __UTIL_H__ 1

unsigned long crc32(unsigned char *buf, unsigned int len);

void fatal (char *str, ...);
void error (char *str, ...);
void warn (char *str, ...);
void info (char *str, ...);
void debug (char *str, ...);
int delim (char c);
char *skip_white (char *ptr);
void eword (int wd);
void emit (int byte);
void f_record (void);
void hexout (int byte);
void binout (int byte);
void print_line (void);
int head (char *str1, char *str2);

#ifndef __GNUC__
/* GCC will use the inline versions here in the util.h header file,
   and other compilers will use the regular versions in util.c */

static int any (char c, char *str) {
static char mapdn (char c) {
static int lobyte (int i) {
static int hibyte (int i) {
static int alpha (char c) {
static int alphan (char c) {
static int white (char c) {
#else

  /* I've replaced most of the original code with more portable current library calls */
  
/*
 *      any --- does str contain c?
 */
static inline int any (char c, char *str) {
  if (c == EOS) return NO;
  return strchr(str, c) != NULL;
}

/*
 *      mapdn --- convert A-Z to a-z
 */
static inline char mapdn (char c) {
  if (isalpha(c) && isupper(c)) return tolower(c);
  return (c);
}

/*
 *      lobyte --- return low byte of an int
 */
static inline int lobyte (int i) {
  return (i & 0xFF);
}

/*
 *      hibyte --- return high byte of an int
 */
static inline int hibyte (int i) {
  return ((i >> 8) & 0xFF);
}

/*
 *      alpha --- is character a legal letter
 */
static inline int alpha (char c) {
  return isalpha(c) || (c == '_') || (c == '.');
}

/*
 *      alphan --- is character a legal letter or digit
 */
static inline int alphan (char c) {
  return isalnum(c) || (c == '_') || (c == '.') || (c == '$');
}

/*
 *      white --- is character whitespace?
 */
static inline int white (char c) {
  /* might later switch to either isblank() or strtok(), after checking for unintended consequences... */
  return any (c, " \n\t\r");
}

#endif /* __GNUC__ */

#endif
